"""Implementation of postgres backends."""

from dinao.backend.base import Connection, ConnectionPool
from dinao.backend.errors import BackendNotInstalledError, ConfigurationError


class ConnectionPSQLPsycopg2(Connection):
    """Implementation of Connection for Psycopg2."""

    def _execute(self, cursor, sql: str, params: tuple = None):
        cursor.execute(query=sql, vars=params)


class ConnectionPoolPSQLPsycopg2(ConnectionPool):
    """Implementation of ConnectionPool for Psycopg2."""

    def __init__(self, db_url: str):
        """Construct a connection pool for the given connection URL.

        The db_url is expected to be in the following format::

            "postgres+psycopg2://{username}:{password}@{hostname}:{port}/{db_name}?{optional_args}"

        Supported `optional_args` include:

            * schema, a list of strings that sets the search path of the connections, defaults to "public"
            * pool_min_conn, an integer specifying the minimum connections to keep in the pool, defaults to 1
            * pool_max_conn, an integer specifying the maximum connections to keep in the pool, defaults to 1
            * pool_threaded, a boolean specifying a threaded pool should be used, defaults to False
            * sslmode, a string ("prefer", "verify-full", etc ...) specifying ssl mode for connections, defaults to None
            * sslrootcert, a string specifying a path to a root CA to use for ssl verification, defaults to None

        :param db_url: a url with the described format
        :raises: ConfigurationError, BackendNotInstalledError
        """
        super().__init__(db_url)
        try:
            import psycopg2.pool
        except ModuleNotFoundError:  # pragma: no cover
            issue = "Module psycopg2 not installed, cannot create connection pool"
            raise BackendNotInstalledError(issue)
        self._cnx_kwargs = self._make_cnx_kwargs()
        self._pool_class = psycopg2.pool.SimpleConnectionPool
        if self._get_arg("pool_threaded", bool, False):
            self._pool_class = psycopg2.pool.ThreadedConnectionPool
        self._pool_impl = None
        self._raise_for_unexpected_args()
        self._pool = self._pool_class(**self._cnx_kwargs)

    def _make_cnx_kwargs(self):
        dbname = self._db_url.path.strip("/")
        if not dbname:
            raise ConfigurationError("Database name is required but missing")
        schema = ",".join(self._get_arg("schema", list, ["public"]))
        min_c = self._get_arg("pool_min_conn", int, 1)
        max_c = self._get_arg("pool_max_conn", int, min_c)
        if min_c <= 0 or max_c <= 0:
            raise ConfigurationError("The pool_max_conn and pool_min_conn must be greater than 0")
        if max_c < min_c:
            raise ConfigurationError("The argument pool_max_conn must be greater or equal to pool_min_conn")
        ssl_mode = self._get_arg("sslmode", str, None)
        ssl_root_cert = self._get_arg("sslrootcert", str, None)
        kwargs = {
            "maxconn": max_c,
            "minconn": min_c,
            "dbname": dbname,
            "user": self._db_url.username,
            "password": self._db_url.password,
            "host": self._db_url.hostname,
            "port": self._db_url.port,
            "options": f"-c search_path={schema}",
        }
        if ssl_mode:
            kwargs["sslmode"] = ssl_mode
        if ssl_root_cert:
            kwargs["sslrootcert"] = ssl_root_cert
        return kwargs

    def lease(self) -> Connection:  # noqa: D102
        inner_cnx = self._pool.getconn()
        return ConnectionPSQLPsycopg2(inner_cnx)

    def release(self, cnx: Connection):  # noqa: D102
        self._pool.putconn(cnx._cnx)

    def dispose(self):  # noqa: D102
        if not self._pool.closed:
            self._pool.closeall()

    @property
    def mung_symbol(self) -> str:  # noqa: D102
        return "%s"



"""Implements a simple templating grammar / parser for binding SQL to functions."""

from typing import Any, Tuple

from dinao.binding.errors import TemplateError

# fmt: off
from pyparsing import (  # noqa: I101
    alphanums, alphas, printables,
    Combine, Forward, Group, OneOrMore, Suppress, White, Word, ZeroOrMore, ParseBaseException,
    ParseResults
)
# fmt: on


class TemplateParameter:
    """Represents a parameter for prepared statement rendered from a template."""

    def __init__(self, kwarg_path: Tuple[str, ...], is_replace: bool = False):
        """Construct a template parameter.

        :param kwarg_path: an immutable list representing the 'path' of the parameter in a root dictionary of arguments
        :param is_replace: is this parameter a direct replacement
        """
        self.is_replace = is_replace
        self.kwarg_path = kwarg_path


def set_result_type(subject: str, position: int, result: ParseResults):
    """For use with forward lookup groups to inject suppressed characters into the parse result.

    :param subject: string being parsed
    :param position: position in the string the parsing is taking place
    :param result: the parse result object for the given position / subject
    """
    result[0].insert(0, subject[position])


class Template:
    """Implements templating supporting only argument replacement."""

    # Grammar is based roughly on problem description from
    # https://github.com/pyparsing/pyparsing/issues/5
    # fmt: off
    OPEN_VAR     = Suppress("#{")                                                           # noqa: E221
    OPEN_REP     = Suppress("!{")                                                           # noqa: E221
    OPENS        = (OPEN_VAR | OPEN_REP)                                                    # noqa: E221
    CLOSE        = Suppress("}")                                                            # noqa: E221
    LONERS       = ((~OPENS + "#") | (~OPENS + "!") |                                       # noqa: E221
                    (~OPENS + "{") | (~OPENS + "}"))                                        # noqa: E221
    IDENTIFIER   = Word(alphas, alphanums + "_")                                            # noqa: E221
    REPLACEMENT  = (IDENTIFIER + ZeroOrMore(Suppress(".") + IDENTIFIER))                    # noqa: E221
    PLAIN_TEXT   = Word(printables, exclude_chars="#!{}")                                   # noqa: E221
    NULL_SPACE   = ZeroOrMore(White())                                                      # noqa: E221
    LOOKUP_VAR   = Forward()                                                                # noqa: E221
    LOOKUP_REP   = Forward()                                                                # noqa: E221
    SQL_FRAGMENT = Combine(NULL_SPACE + OneOrMore(PLAIN_TEXT | LONERS) + NULL_SPACE)        # noqa: E221
    LOOKUP_VAR  << Group(OPEN_VAR + REPLACEMENT + CLOSE).add_parse_action(set_result_type)  # noqa: E221
    LOOKUP_REP  << Group(OPEN_REP + REPLACEMENT + CLOSE).add_parse_action(set_result_type)  # noqa: E221
    LOOKUP       = (LOOKUP_VAR | LOOKUP_REP)                                                # noqa: E221
    GRAMMAR      = OneOrMore(NULL_SPACE + LOOKUP + NULL_SPACE | SQL_FRAGMENT)               # noqa: E221
    # fmt: on

    def __init__(self, sql_template: str):
        """Construct a simple variable replacement template.

        :param sql_template: the raw SQL template before parsing as a plain string.
        """
        self._sql_template = sql_template
        self._parsed_template = []
        self._arguments = []
        try:
            nodes = self.GRAMMAR.parse_string(sql_template, parse_all=True)
        except ParseBaseException as x:
            raise TemplateError(f"{x.msg}:\n{x.line}\n{(' ' * (x.col - 1))}^")
        for node in nodes:
            if not isinstance(node, str):
                is_replace = node.pop(0) == "!"
                node = tuple(map(str, node))
                self._arguments.append(node)
                self._parsed_template.append(TemplateParameter(node, is_replace))
                continue
            # If the previous element was a string, concat it with the new str node otherwise append it
            previous = self._parsed_template.pop() if self._parsed_template else ""
            previous = [previous + node] if isinstance(previous, str) else [previous, node]
            self._parsed_template += previous
        self._arguments = tuple(self._arguments)

    @property
    def arguments(self) -> tuple[Any, ...]:
        """Return the arguments expected by the template."""
        return self._arguments

    def __str__(self) -> str:
        """Return a simple representation of the template."""
        return self._sql_template

    @staticmethod
    def _resolve_value(kwarg_path: Tuple[str, ...], root_args: dict):
        node = root_args
        for arg_name in kwarg_path:
            node = node[arg_name] if isinstance(node, dict) else getattr(node, arg_name)
        return node

    def render(self, mung_symbol: str, kwargs: dict) -> tuple[str | Any, tuple[Any, ...]]:
        """Render the template to SQL execution arguments.

        :param mung_symbol: the symbol used to replace parameters in the template
        :param kwargs: the root dictionary of key word arguments to resolve parameters from

        :returns: a tuple where the first element is a SQL statement and the second is a tuple of its parameters
        """
        munged = ""
        parameters = []
        cache = {}
        for frag in self._parsed_template:
            if isinstance(frag, TemplateParameter):
                value = cache.get(frag.kwarg_path) or self._resolve_value(frag.kwarg_path, kwargs)
                cache[frag.kwarg_path] = value
                if frag.is_replace:
                    frag = str(value)
                else:
                    parameters.append(value)
                    frag = mung_symbol
            munged += frag
        return munged, tuple(parameters)



"""Defines common errors raised from database backend code."""


class UnsupportedBackendError(Exception):
    """Raised when an unsupported backend is specified."""

    pass


class ConfigurationError(Exception):
    """Raised when there is a backend configuration connection error."""

    pass


class BackendNotInstalledError(Exception):
    """Raised when a backend engine is not installed."""

    pass


class ConnectionPoolClosed(Exception):
    """Raised when a connection pool is closed."""

    pass



"""Implementation of mysql backends."""

import uuid
from contextlib import contextmanager
from typing import Any, Generator

from dinao.backend.base import Connection, ConnectionPool, ResultSet
from dinao.backend.errors import BackendNotInstalledError, ConfigurationError, ConnectionPoolClosed


class ConnectionMySQL(Connection):
    """Implementation of Connection for MySQL Connector."""

    def _execute(self, cursor, sql: str, params: tuple = None):
        cursor.execute(sql, params)

    @contextmanager
    def query(self, sql: str, params: tuple = None) -> Generator[ResultSet, Any, None]:
        """Execute the given SQL as a statement with the given parameters. Provide the results as context.

        :param sql: the SQL statement(s) to execute
        :param params: the values to bind to the execution of the given SQL
        :returns: a result set representing the query's results
        """
        cursor = self._cnx.cursor()
        self._execute(cursor, sql, params)
        try:
            yield ResultSet(cursor)
        finally:
            cursor.reset()
            cursor.close()


class ConnectionPoolMySQL(ConnectionPool):
    """Implementation of ConnectionPool for MySQL Connector."""

    def __init__(self, db_url: str):
        """Construct a connection pool for the given connection URL.

        The db_url is expected to be in the following format::

            "mysql://{username}:{password}@{hostname}:{port}/{db_name}?{optional_args}"

        Supported `optional_args` include:

            * pool_name, a string specifying a unique name for the pool, defaults to a random string
            * pool_size, an integer specifying the size of the pool, defaults to 5
            * ssl_ca, an absolute path to a certificate to verify the server with, defaults to None
            * ssl_verify_cert, a boolean specifying that the server's cert must be verified, defaults to None

        :param db_url: a url with the described format
        :raises: ConfigurationError, BackendNotInstalledError
        """
        super().__init__(db_url)
        try:
            from mysql.connector.pooling import MySQLConnectionPool
        except ModuleNotFoundError:  # pragma: no cover
            issue = "Module mariadb not installed, cannot create connection pool"
            raise BackendNotInstalledError(issue)
        self._pool_kwargs = self._make_cnx_kwargs()
        self._raise_for_unexpected_args()
        self._pool = MySQLConnectionPool(**self._pool_kwargs)
        self._closed = False

    def _make_cnx_kwargs(self) -> dict:
        dbname = self._db_url.path.strip("/")
        if not dbname:
            raise ConfigurationError("Database name is required but missing")
        pool_name = self._get_arg("pool_name", str, f"dinao-pool-{uuid.uuid4()}")
        if not pool_name:
            raise ConfigurationError("The value for pool_name cannot be an empty string")
        pool_size = self._get_arg("pool_size", int, 5)
        if pool_size <= 0:
            raise ConfigurationError("The value for pool_size must be greater than 0")
        kwargs = {
            "database": dbname,
            "pool_name": pool_name,
            "pool_size": pool_size,
            "user": self._db_url.username,
            "password": self._db_url.password,
            "host": self._db_url.hostname,
            "port": self._db_url.port,
        }
        ssl_ca = self._get_arg("ssl_ca", str, None)
        ssl_verify_cert = self._get_arg("ssl_verify_cert", bool, None)
        if ssl_ca is not None:
            kwargs["ssl_ca"] = ssl_ca
        if ssl_verify_cert is not None:
            kwargs["ssl_verify_cert"] = ssl_verify_cert
        return kwargs

    def lease(self) -> Connection:  # noqa: D102
        if self._closed:
            raise ConnectionPoolClosed("Pool is closed")
        inner_cnx = self._pool.get_connection()
        return ConnectionMySQL(inner_cnx)

    def release(self, cnx: Connection):  # noqa: D102
        cnx._cnx.close()

    def dispose(self):  # noqa: D102
        if not self._closed:
            self._closed = True

    @property
    def mung_symbol(self) -> str:  # noqa: D102
        return "%s"



"""Mapper interface definition and stock implementations."""

import collections.abc
import inspect
import typing
from abc import ABC, abstractmethod
from datetime import datetime
from uuid import UUID

from dinao.backend.base import ColumnDescriptor
from dinao.binding.errors import TooManyValuesError

TUPLE_GENERICS = [tuple, typing.Tuple]
GENERATOR_GENERICS = [collections.abc.Generator, typing.Generator]
DICT_GENERICS = [dict, typing.Dict, typing.Mapping, collections.abc.Mapping]
LIST_GENERICS = [list, typing.List, typing.Iterable, collections.abc.Sequence, collections.abc.Iterable]
NATIVE_SINGLE = [str, int, float, complex, bool, datetime, UUID]

# Typing special forms that should not be treated as mappable classes
# In Python 3.14+, inspect.isclass() returns True for these, but they cannot be instantiated
TYPING_SPECIAL_FORMS = {typing.Union, typing.Optional}


class RowMapper(ABC):
    """Interface for database results mapper."""

    @abstractmethod
    def __call__(self, row: typing.Tuple, description: typing.Tuple[ColumnDescriptor, ...]):
        """Map a database row to the type this implementation maps to.

        :param row: a tuple of raw values from a database result set
        :param description: the description of the row's columns

        :returns: the row as the mapped type.
        """
        pass  # pragma: no cover


class TupleRowMapper(RowMapper):
    """Implements mapping for simply returning the raw result tuple with no casting."""

    def __call__(self, row: typing.Tuple, description: typing.Tuple[ColumnDescriptor, ...]):  # noqa: D102
        return row


class DictRowMapper(RowMapper):
    """Implements mapping for dictionaries where keys are string and values are not cast."""

    def __call__(self, row: typing.Tuple, description: typing.Tuple[ColumnDescriptor, ...]):  # noqa: D102
        return {d.name: row[col] for col, d in enumerate(description)}


class ClassRowMapper(DictRowMapper):
    """Implements mapping for classes where rows are passed as kwargs."""

    def __init__(self, mapped_class):
        """Construct a class row mapper by passing the row as named key word args to the class constructor.

        :param mapped_class: the class to map a row to
        """
        self._mapped_class = mapped_class

    def __call__(self, row: typing.Tuple, description: typing.Tuple[ColumnDescriptor, ...]):  # noqa: D102
        kwargs = super().__call__(row, description)
        return self._mapped_class(**kwargs)


class SingleValueRowMapper(RowMapper):
    """Implements a row mapper for a primitive type."""

    def __call__(self, row: typing.Tuple, description: typing.Tuple[ColumnDescriptor, ...]):  # noqa: D102
        if len(row) > 1:
            raise TooManyValuesError(f"Too many values, expected 1, got {len(row)}")
        return row[0]


def get_row_mapper(row_type: typing.Type) -> typing.Optional[RowMapper]:
    """Find the row mapper for the given row type.

    :param row_type: the type to find a mapper for

    :returns: a row mapper if one could be determined, None otherwise
    """
    generics_type = typing.get_origin(row_type)
    generics_args = typing.get_args(row_type)
    # Currently we don't support things like Tuple[str, int, ... ], Dict[str, int] etc ...
    if generics_type and generics_args:
        return None
    elif row_type in TUPLE_GENERICS:
        return TupleRowMapper()
    elif row_type in DICT_GENERICS:
        return DictRowMapper()
    elif row_type in NATIVE_SINGLE:
        return SingleValueRowMapper()
    # Exclude typing special forms that look like classes in Python 3.14+
    # Bare Union/Optional without type parameters cannot be mapped
    elif row_type in TYPING_SPECIAL_FORMS:
        return None
    # Finally fall back to kwargs init
    elif inspect.isclass(row_type):
        return ClassRowMapper(row_type)
    # Signal we don't know what to do
    return None



"""Defines common errors raised from function binding."""


class NoPoolSetError(Exception):
    """Raised when there is no pool set for a function binder before a sql operation is called."""

    pass


class PoolAlreadySetError(Exception):
    """Raised when a pool is set twice on a binder."""

    pass


class BindingError(Exception):
    """Base exception for errors when binding functions."""

    pass


class MappingError(Exception):
    """Base exception for errors related to mapping database results to return types."""

    pass


class TemplateError(BindingError):
    """Raised when parsing a template fails."""

    pass


class FunctionAlreadyBoundError(BindingError):
    """Raised when a function is bound more than once to a query, execution, or transaction."""

    pass


class SignatureError(BindingError):
    """Base exception for binding errors related to function signature inspection."""

    pass


class BadReturnTypeError(SignatureError):
    """Raised when a return hint specifies a type that cannot be used for mapping in the context of the binding."""

    pass


class MissingTemplateArgumentError(SignatureError):
    """Raised when a template specifies an argument not found in its bounded function's signature."""

    pass


class CannotInferMappingError(SignatureError):
    """Raised when the return mapping for a bound function cannot be determined."""

    pass


class MultipleConnectionArgumentError(SignatureError):
    """Raised when a bound function specifies multiple connection arguments."""

    pass


class TooManyValuesError(MappingError):
    """Raised when the number of columns does not match the expected number for mapping."""

    pass


class TooManyRowsError(MappingError):
    """Raised when mapping implies a singular return, but many rows are returned."""

    pass



"""Defines a basic and primitive database interface. It is basically a thin wrapper on DB API 2.0."""

import logging
from abc import ABC, abstractmethod
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Any, Generator, List, Tuple
from urllib.parse import parse_qs, urlparse

from dinao.backend.errors import ConfigurationError


@dataclass
class ColumnDescriptor:
    """Describes a column in a result set."""

    name: str
    type_code: int
    display_size: int = None
    internal_size: int = None
    precision: int = None
    scale: int = None
    null_ok: bool = None


class ResultSet:
    """Basic interface definition for result sets (a.k.a rows) returned from Database queries."""

    def __init__(self, cursor):
        """Construct a result set.

        :param cursor: the underlying DB API 2.0 cursor being wrapped by this object.
        """
        self._cursor = cursor
        self._columns = None
        self._description = None

    def fetchone(self) -> Tuple:
        """Fetch one result tuple from the underlying cursor.

        If no results are left, None is returned.

        :returns: a tuple representing a result row or None
        """
        return self._cursor.fetchone()

    def fetchall(self) -> List[Tuple]:
        """Fetch the *remaining* result tuples from the underlying cursor.

        If no results are left, an empty list is returned.

        :returns: a list of tuples that are the remaining results of the underlying cursor.
        """
        return self._cursor.fetchall()

    @property
    def description(self) -> Tuple[ColumnDescriptor]:
        """Return a sequence of column descriptions representing the result set.

        :returns: a tuple of ColumnDescriptors
        """
        if not self._description:
            self._description = tuple([ColumnDescriptor(*(d[0:7])) for d in self._cursor.description])
        return self._description

    @property
    def rowcount(self) -> int:
        """Return the row count of the result set.

        :returns: the integer count of the rows in the result set
        """
        return self._cursor.rowcount


class Connection(ABC):
    """Basic interface definition for a database connection."""

    def __init__(self, cnx, auto_commit: bool = True):
        """Construct a Connection object.

        :param cnx: the inner DB API 2.0 connection this object wraps
        :param auto_commit: should calls to execute() be automatically committed, defaults to True
        """
        self.logger = logging.getLogger(__name__)
        self._cnx = cnx
        self._auto_commit = auto_commit

    @property
    def autocommit(self):
        """Whether commit is called after every call to query(...) and execute(...)."""
        return self._auto_commit

    @autocommit.setter
    def autocommit(self, value: bool):
        self._auto_commit = value

    def commit(self):
        """Commit changes for this connection / transaction to the database."""
        self._cnx.commit()

    def rollback(self):
        """Rollback changes for this connection / transaction to the database."""
        self._cnx.rollback()

    @abstractmethod
    def _execute(self, cursor, sql: str, params: tuple = None):
        pass  # pragma: no cover

    @contextmanager
    def query(self, sql: str, params: tuple = None) -> Generator[ResultSet, Any, None]:
        """Execute the given SQL as a statement with the given parameters. Provide the results as context.

        :param sql: the SQL statement(s) to execute
        :param params: the values to bind to the execution of the given SQL
        :returns: a result set representing the query's results
        """
        cursor = self._cnx.cursor()
        self._execute(cursor, sql, params)
        try:
            yield ResultSet(cursor)
        finally:
            cursor.close()

    def execute(self, sql: str, params: tuple = None, commit: bool = None) -> int:
        """Execute the given SQL as a statement with the given parameters and return the affected row count.

        :param sql: the SQL statement(s) to execute
        :param params: the values to bind to the execution of the given SQL
        :param commit: commit the changes to the database after execution, defaults to value given in constructor
        """
        commit = commit if commit is not None else self._auto_commit
        cursor = self._cnx.cursor()
        self._execute(cursor, sql, params)
        affected = cursor.rowcount
        if commit:
            self.commit()
        cursor.close()
        return affected


class ConnectionPool(ABC):
    """Basic interface definition for a pool of database connections."""

    def __init__(self, db_url: str):
        """Construct a connection pool for the given connection URL.

        The db_url is expected to be in the following format::

            "{dialect}+{driver}://{username}:{password}@{hostname}:{port}/{db_name}?{optional_args}"

        :param db_url: a url with the described format
        """
        self.logger = logging.getLogger(__name__)
        self._raw_db_url = db_url
        self._db_url = urlparse(self._raw_db_url)
        self._args = parse_qs(self._db_url.query, keep_blank_values=True)

    @staticmethod
    def _strict_bool(value: str):
        if value.lower() not in ["true", "false"]:
            raise ValueError(f"Cannot cast '{value}' to bool")
        return value.lower() == "true"

    def _raise_for_unexpected_args(self):
        unexpected = ",".join(self._args.keys())
        if unexpected:
            raise ConfigurationError(f"Unexpected argument(s): {unexpected}")

    def _get_arg(self, name: str, expected_type, default=None):
        if name not in self._args:
            self.logger.debug(f"No '{name}' specified, defaulting to {default}")
            return default
        caster = expected_type if expected_type is not bool else self._strict_bool
        try:
            if caster != list:
                if len(self._args.get(name)) != 1:
                    raise ConfigurationError(f"Invalid argument '{name}': only a single value must be specified")
                return caster(self._args.pop(name)[0])
            return self._args.pop(name)
        except ValueError as x:
            raise ConfigurationError(f"Invalid argument '{name}': must be {expected_type.__name__}") from x

    @property
    @abstractmethod
    def mung_symbol(self) -> str:
        """Return the symbol used when replacing variable specifiers in templated SQL."""
        pass  # pragma: no cover

    @abstractmethod
    def lease(self) -> Connection:
        """Lease a connection from the underlying pool."""
        pass  # pragma: no cover

    @abstractmethod
    def release(self, cnx: Connection):
        """Release a connection back to the underlying pool."""
        pass  # pragma: no cover

    @abstractmethod
    def dispose(self):
        """Close the pool and clean up any resources it was using."""
        pass  # pragma: no cover



"""Implementation of SQLite backends."""

import os.path
import sqlite3

from dinao.backend.base import Connection, ConnectionPool


class ConnectionSQLite3(Connection):
    """Implementation of Connection for Sqlite3."""

    def _execute(self, cursor, sql: str, params: tuple = None):
        if params:
            cursor.execute(sql, params)
            return
        cursor.execute(sql)


class ConnectionPoolSQLite3(ConnectionPool):
    """Implementation of ConnectionPool for SQLite3."""

    def __init__(self, db_url: str):
        """Construct a connection pool for the given connection URL.

        The db_url is expected to be in the following format::

            "sqlite3://{filename}"

        Obviously, this is not a real "pool", all connections given out by lease will be new connections.

        :param db_url: a url with the described format
        :raises: ConfigurationError, BackendNotInstalledError
        """
        super().__init__(db_url)
        self._cnx_kwargs = self._url_to_cnx_kwargs()
        self._raise_for_unexpected_args()

    def _url_to_cnx_kwargs(self):
        file_path = os.path.abspath(os.path.expanduser(self._db_url.path))
        return {"database": file_path}

    def lease(self) -> Connection:  # noqa: D102
        inner_cnx = sqlite3.connect(**self._cnx_kwargs)
        return ConnectionSQLite3(inner_cnx)

    def release(self, cnx: Connection):  # noqa: D102
        cnx._cnx.close()

    def dispose(self):  # noqa: D102
        return

    @property
    def mung_symbol(self) -> str:  # noqa: D102
        return "?"



"""Functionality abstracting the primitive database backend interface."""

from urllib.parse import urlparse

from dinao.backend.base import Connection, ConnectionPool, ResultSet
from dinao.backend.errors import ConfigurationError, UnsupportedBackendError
from dinao.backend.mariadb import ConnectionPoolMariaDB
from dinao.backend.mysql import ConnectionPoolMySQL
from dinao.backend.postgres import ConnectionPoolPSQLPsycopg2
from dinao.backend.sqlite import ConnectionPoolSQLite3

ENGINE_DEFAULTS = {"postgresql": "psycopg2", "sqlite3": None, "mariadb": "mariadbconnector", "mysql": "mysqlconnector"}


def create_connection_pool(db_url: str) -> ConnectionPool:
    """Create a connection pool for the given database connection URL.

    The db_url is expected to be in the following format::

        "{db_backend}+{driver}://{username}:{password}@{hostname}:{port}/{db_name}"

    With different db_backends / drivers supporting additional arguments.

    :returns: A connection pool based on the given database URL.
    :raises: ConfigurationError, UnsupportedBackendError
    """
    parsed_url = urlparse(db_url)
    backend = parsed_url.scheme
    if not backend:
        raise ConfigurationError("No database backend specified")
    backend = backend.split("+")
    engine = ENGINE_DEFAULTS.get(backend[0]) if len(backend) == 1 else backend[1]
    backend = backend[0]
    if backend == "postgresql" and engine == "psycopg2":
        return ConnectionPoolPSQLPsycopg2(db_url)
    if backend == "sqlite3" and engine is None:
        return ConnectionPoolSQLite3(db_url)
    if backend == "mariadb" and engine == "mariadbconnector":
        return ConnectionPoolMariaDB(db_url)
    if backend == "mysql" and engine == "mysqlconnector":
        return ConnectionPoolMySQL(db_url)
    raise UnsupportedBackendError(f"The backend+engine '{parsed_url.scheme}' is not supported")


__all__ = ["Connection", "ConnectionPool", "ResultSet", "create_connection_pool", "errors"]



"""Implements functionality for binding python functions to SQL queries and actions."""

import inspect
import threading
import types
import typing
from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import Any, Generator

from dinao.backend.base import Connection, ConnectionPool, ResultSet
from dinao.binding.errors import (
    BadReturnTypeError,
    CannotInferMappingError,
    FunctionAlreadyBoundError,
    MissingTemplateArgumentError,
    MultipleConnectionArgumentError,
    NoPoolSetError,
    PoolAlreadySetError,
    TemplateError,
    TooManyRowsError,
)
from dinao.binding.mappers import DICT_GENERICS, GENERATOR_GENERICS, LIST_GENERICS, RowMapper, get_row_mapper
from dinao.binding.templating import Template


def _unwrap_optional(type_hint: typing.Any) -> typing.Optional[typing.Type]:
    """Extract the inner type from Optional[X] or X | None.

    :param type_hint: A type hint that may be Optional[X] or X | None
    :returns: The inner type X if type_hint is Optional[X], None otherwise
    """
    origin = typing.get_origin(type_hint)
    args = typing.get_args(type_hint)
    # Check for Union types (typing.Union or types.UnionType for X | Y syntax)
    if origin not in (typing.Union, types.UnionType) or not args:
        return None
    # Filter out NoneType from the union args
    non_none_args = [a for a in args if a is not type(None)]
    # Only unwrap if there's exactly one non-None type (true Optional)
    if len(non_none_args) == 1:
        return non_none_args[0]
    return None


class FunctionBinder:
    """Implements functionality for binding functions to SQL queries and actions."""

    def __init__(self):
        """Construct a function binder."""
        self._context_store = threading.local()
        self._verbose_trace = False
        self._cnx_pool = None

    def _suppressed_raise(self, exc: Exception):
        if self._verbose_trace:
            raise exc  # pragma: no cover
        # Try to cut down on the traces so the user gets closer to the issue in their code
        raise exc.with_traceback(None) from exc

    def _raise_for_multi_binding(self, func: callable):
        if isinstance(func, BoundedFunction):
            name = func.bounded_function.__name__
            error = f"The function {name} has already been bounded by {func}"
            self._suppressed_raise(FunctionAlreadyBoundError(error))

    def _make_template(self, sql: str):
        try:
            return Template(sql)
        except TemplateError as exc:
            self._suppressed_raise(exc)

    @property
    def pool(self) -> ConnectionPool:
        """Get the connection pool used by this binder."""
        return self._cnx_pool

    @pool.setter
    def pool(self, pool: ConnectionPool):
        """Set the connection pool used by this binder.

        :raises PoolAlreadySetError
        """
        if self._cnx_pool is not None:
            raise PoolAlreadySetError("The connection pool can only be set once")
        self._cnx_pool = pool

    @property
    def mung_symbol(self) -> str:
        """Return the mung symbol used for rendering templates bound by the binder."""
        if self._cnx_pool is None:
            raise NoPoolSetError("No connection pool has been set for the binder")
        return self.pool.mung_symbol

    def execute(self, sql: str) -> callable:
        """Binds a given function to a given SQL template.

        .. warning::
            The execution of the SQL given is not expected to return results / rows. Use the

        .. note::
            The return type of the decorated / bound function may return an int, which will make it return the number of
            rows affected by the execution of the last SQL statement in the template.

        Example::

            @binder.execute("INSERT INTO my_table (col1, col2) VALUES (#{arg1.some_key_or_member}, #{arg2})")
            def my_bounded_func(arg1: dict, arg2: str) -> int:
                pass

        :param sql: a SQL template to bind the function execution to
        :returns: a decorator expecting a callable
        :raises: BadReturnTypeError, FunctionAlreadyBoundError
        """
        # fmt: off
        def decorated(func: callable):
            self._raise_for_multi_binding(func)
            template = self._make_template(sql)
            return BoundedExecution(self, template, func)
        return decorated
        # fmt: on

    def query(self, sql: str) -> callable:
        """Binds a given function to a given SQL template.

        .. warning::
            Currently return type mapping is not implemented if the return type specified on the bounded function is not
            either empty or None a NotImplementedError is raised.  A list of tuples is returned when the return hint is
            left empty, representing the results from the query.

        Example::

            @binder.query("SELECT * INTO my_table WHERE col_1 = #{arg1.some_key_or_member} AND col_2 = #{arg2}")
            def my_bounded_func(arg1: dict, arg2: str):
                pass

        :param sql: a SQL template to bind the function execution to
        :returns: a decorator expecting a callable
        :raises: NotImplementedError, FunctionAlreadyBoundError
        """
        # fmt: off
        def decorated(func: callable):
            self._raise_for_multi_binding(func)
            template = self._make_template(sql)
            return_type = typing.get_origin(inspect.signature(func).return_annotation)
            if return_type in GENERATOR_GENERICS:
                return BoundedGeneratingQuery(self, template, func)
            return BoundedQuery(self, template, func)
        return decorated
        # fmt: on

    def transaction(self):
        """Binds a given function to a given SQL transaction.

        All bounded function called during the call of a function bounded in this way will use the same transaction /
        connection.

        Example::

            @binder.query("SELECT * INTO my_table WHERE col_1 = #{arg1.some_key_or_member} AND col_2 = #{arg2}")
            def my_select(arg1: dict, arg2: str):
                pass

            @binder.execute(
                "INSERT INTO stats_table VALUES "
                "(#{stats.name}, #{stats.value}) "
                "ON CONFLICT (name) DO UPDATE SET value = #{stats.value}"
            )
            def my_update(stat: dict) -> int:
                pass

            @binder.transaction()
            def my_transaction(arg1: dict, stat_name: str) -> int:
                my_table_results = my_select(arg1, stat_name)
                stat = {"name": stat, "value": 0}
                for row in my_table_results:
                    stat["value"] += row[0]
                return my_update(stat=stat)

        :returns: a decorator expecting a callable
        :raises: NotImplementedError, FunctionAlreadyBoundError
        """
        # fmt: off
        def decorated(func: callable):
            self._raise_for_multi_binding(func)
            return BoundedTransaction(self, func)
        return decorated
        # fmt: on

    @contextmanager
    def connection(self) -> Generator[Connection | None | Any, Any, None]:
        """Context manger for database connections used by bound functions.

        An active connection will be kept and yielded, provided to all contexts within the context that initially asked
        for it.

        Example::

            with binder.connection() as cnx_outer:
                with binder.connection() as cnx_inner:
                    assert cnx_outer == cnx_inner  # This assertion is true

        Any exception caught during the context of a connection will trigger a rollback of the transaction.  The
        transaction is automatically committed when execution is yielded back.
        """
        if self._cnx_pool is None:
            raise NoPoolSetError("No connection pool has been set for the binder")
        active_cnx = getattr(self._context_store, "active_cnx", None)
        if active_cnx:
            yield active_cnx
            return
        self._context_store.active_cnx = self.pool.lease()
        try:
            yield self._context_store.active_cnx
            self._context_store.active_cnx.commit()
        except:  # noqa: E722
            self._context_store.active_cnx.rollback()
            raise
        finally:
            self.pool.release(self._context_store.active_cnx)
            self._context_store.active_cnx = None


class BoundedFunction(ABC):
    """Abstract class for common behaviors between bound functions."""

    def __init__(self, binder: FunctionBinder, func: callable):  # noqa: D107
        self._binder = binder
        self._func = func
        self._sig = inspect.signature(func)

    @abstractmethod
    def __call__(self, *args, **kwargs):  # noqa: D102
        pass  # pragma: no cover

    @property
    def bounded_function(self) -> callable:
        """Return the original callable this object binds."""
        return self._func


class BoundedSQLFunction(BoundedFunction):
    """Abstract class for common behaviors between functions bound to templated SQL."""

    def __init__(self, binder: FunctionBinder, sql_template: Template, func: callable):  # noqa: D107
        super().__init__(binder, func)
        self._sql_template = sql_template
        for arg_id in sql_template.arguments:
            if arg_id[0] not in self._sig.parameters:
                error = f"Argument '{arg_id[0]}' specified in template but is not an argument of {func.__name__}"
                raise MissingTemplateArgumentError(error)

    def __str__(self) -> str:
        """Return a simple representation of a SQL bound function."""
        return f"{self.__class__.__name__} of {self._sql_template} ({self.bounded_function.__name__})"

    @abstractmethod
    def __call__(self, *args, **kwargs):  # noqa: D102
        pass  # pragma: no cover


class BoundedQuery(BoundedSQLFunction):
    """Implementation of a bounded function that represents a SQL query that returns rows."""

    def __init__(self, binder: FunctionBinder, sql_template: Template, func: callable, row_mapper: RowMapper = None):
        """Construct a BoundSQLFunction that returns result sets.

        :param binder: the binder to which this bound function belongs
        :param sql_template: the template to use for resolving SQL parameters / munged sql.
        :param func: the callable being bound to the given SQL
        :param row_mapper: a RowMapper implementation that overrides the mapper inferred from the return annotation

        :raises CannotInferMappingError
        """
        super().__init__(binder, sql_template, func)
        self._row_mapper = None
        self._return_impl = None

        return_type = self._sig.return_annotation
        # Default (a.k.a Signature.empty) if no return type given is List[tuple]
        if return_type == inspect.Signature.empty:
            return_type = typing.List[tuple]

        # Handle Optional[X] / X | None by unwrapping to X
        # The _one_return impl already handles None when no row is found
        unwrapped = _unwrap_optional(return_type)
        if unwrapped is not None:
            return_type = unwrapped

        generic_type = typing.get_origin(return_type)
        generic_args = typing.get_args(return_type)
        returns_none = False
        row_type = None

        if return_type in [None, typing.NoReturn]:
            self._return_impl = self._none_return
            returns_none = True
        # Classes / Dataclasses etc ...
        elif return_type and not generic_type:
            self._return_impl = self._one_return
            row_type = return_type
        # Single dictionary returns, no key / value typing suggested
        elif generic_type in DICT_GENERICS and not generic_args:
            self._return_impl = self._one_return
            row_type = dict
        elif generic_type in LIST_GENERICS:
            self._return_impl = self._many_return
            row_type = generic_args[0] if generic_args else tuple
        else:
            raise CannotInferMappingError(f"Unable to determine mapper for {return_type}")

        self._row_mapper = row_mapper or get_row_mapper(row_type)
        if not returns_none and not self._row_mapper:
            raise CannotInferMappingError(f"Unable to determine row mapper for {row_type}")

    @staticmethod
    def _none_return(results: ResultSet):
        return None

    def _one_return(self, results: ResultSet):
        if results.rowcount > 1:
            raise TooManyRowsError(f"Only expected one row, but got {results.rowcount}")
        raw = results.fetchone()
        return self._row_mapper(raw, results.description) if raw else None

    def _many_return(self, results: ResultSet):
        return [self._row_mapper(row, results.description) for row in results.fetchall()]

    def __call__(self, *args, **kwargs):
        """Execute the templated SQL as a query with results.

        :param args: the positional arguments of the bounded / decorated function
        :param kwargs: the keyword arguments of the bounded / decorated function
        """
        bound = self._sig.bind(*args, **kwargs)
        bound.apply_defaults()
        sql, values = self._sql_template.render(self._binder.mung_symbol, bound.arguments)
        with self._binder.connection() as cnx:
            with cnx.query(sql, values) as results:
                return self._return_impl(results)


class BoundedGeneratingQuery(BoundedSQLFunction):
    """Implementation of a bounded function that represents a SQL query that returns rows via a generator."""

    def __init__(self, binder: FunctionBinder, sql_template: Template, func: callable, row_mapper: RowMapper = None):
        """Construct a BoundedGeneratingQuery that returns result sets.

        :param binder: the binder to which this bound function belongs
        :param sql_template: the template to use for resolving SQL parameters / munged sql.
        :param func: the callable being bound to the given SQL
        :param row_mapper: a RowMapper implementation that overrides the mapper inferred from the return annotation

        :raises CannotInferMappingError, BadReturnTypeError
        """
        super().__init__(binder, sql_template, func)
        self._row_mapper = None
        return_type = self._sig.return_annotation
        generic_type = typing.get_origin(return_type)
        generic_args = typing.get_args(return_type)
        if generic_type not in GENERATOR_GENERICS:
            raise BadReturnTypeError(f"Expected results type to be Generator, got {generic_type}")
        # Generators require 3 args but we only support mapping the yield type
        if generic_args and generic_args[1:3] != (type(None), type(None)):
            raise CannotInferMappingError("Only yield_type should be specified for a generator")
        self._row_mapper = row_mapper or get_row_mapper(generic_args[0] if generic_args else tuple)

    def __call__(self, *args, **kwargs):
        """Execute the templated SQL as a query with results.

        :param args: the positional arguments of the bounded / decorated function
        :param kwargs: the keyword arguments of the bounded / decorated function
        """
        bound = self._sig.bind(*args, **kwargs)
        bound.apply_defaults()
        sql, values = self._sql_template.render(self._binder.mung_symbol, bound.arguments)
        with self._binder.connection() as cnx:
            with cnx.query(sql, values) as results:
                row = results.fetchone()
                while row:
                    yield self._row_mapper(row, results.description)
                    row = results.fetchone()


class BoundedExecution(BoundedSQLFunction):
    """Implementation of a bounded function that represents a SQL query that does not return rows."""

    def __init__(self, binder: FunctionBinder, sql_template: Template, func: callable):
        """Construct a BoundSQLFunction that returns no results.

        :param binder: the binder to which this bound function belongs
        :param sql_template: the template to use for resolving SQL parameters / munged sql.
        :param func: the callable being bound to the given SQL

        :raises: BadReturnTypeError
        """
        super().__init__(binder, sql_template, func)
        self._returns_none = False
        if self._sig.return_annotation == self._sig.empty or self._sig.return_annotation is None:
            self._returns_none = True
        elif self._sig.return_annotation != int:
            raise BadReturnTypeError("Bound executions can only return None or int (e.g. affected rows)")

    def __call__(self, *args, **kwargs):
        """Execute the templated SQL as a query without results.

        :param args: the positional arguments of the bounded / decorated function
        :param kwargs: the keyword arguments of the bounded / decorated function
        """
        bound = self._sig.bind(*args, **kwargs)
        bound.apply_defaults()
        sql, values = self._sql_template.render(self._binder.mung_symbol, bound.arguments)
        with self._binder.connection() as cnx:
            affected = cnx.execute(sql, values)
            return None if self._returns_none else affected


class BoundedTransaction(BoundedFunction):
    """Implementation of a bounded function whose call is done within a database transaction."""

    def __init__(self, binder: FunctionBinder, func: callable):
        """Construct a BoundFunction that represents a transaction.

        :param binder: the binder to which this bound function belongs
        :param func: the callable being bound to the given SQL

        :raises: MultipleConnectionArgumentError
        """
        super().__init__(binder, func)
        self._cnx_arg_name = None
        for name in self._sig.parameters:
            if issubclass(self._sig.parameters[name].annotation, Connection):
                if self._cnx_arg_name is not None:
                    error = f"Connection argument specified multiple times for {self.bounded_function.__name__}"
                    raise MultipleConnectionArgumentError(error)
                self._cnx_arg_name = name

    def __str__(self) -> str:
        """Return a simple representation of a transaction bound function."""
        return f"{self.__class__.__name__} ({self.bounded_function.__name__})"

    def __call__(self, *args, **kwargs):
        """Execute the decorated / bound function with a single database connection / transaction.

        :param args: the positional arguments of the bounded / decorated function
        :param kwargs: the keyword arguments of the bounded / decorated function
        """
        with self._binder.connection() as cnx:
            cnx.autocommit = False
            if self._cnx_arg_name:
                kwargs = kwargs or {}
                kwargs[self._cnx_arg_name] = cnx
            return self._func(*args, **kwargs)
